#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import keras
import tensorflow as tf

from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, WeightQuantizer, OutputQuantizer
from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH


# Restrict list of exported symbols on default import
__all__ = ["ClassToken", "QuantizedClassToken",
           "AddPositionEmbs", "QuantizedAddPositionEmbs",
           "ExtractToken", "QuantizedExtractToken"]


@tf.keras.utils.register_keras_serializable()
class ClassToken(keras.layers.Layer):
    """Append a class token to an input layer."""

    def build(self, input_shape):
        super().build(input_shape)
        cls_init = keras.initializers.TruncatedNormal(stddev=0.02)
        self.hidden_size = input_shape[-1]
        self.cls = tf.Variable(
            name="cls",
            initial_value=cls_init(
                shape=(1, 1, self.hidden_size), dtype="float32"),
            trainable=True,
        )

    def call(self, inputs, training=None):
        batch_size = tf.shape(inputs)[0]
        cls_broadcasted = tf.cast(
            tf.broadcast_to(self.cls, [batch_size, 1, self.hidden_size]),
            dtype=inputs.dtype,
        )
        return tf.concat([cls_broadcasted, inputs], 1)


@tf.keras.utils.register_keras_serializable()
class AddPositionEmbs(keras.layers.Layer):
    """Adds (optionally learned) positional embeddings to the inputs."""

    def build(self, input_shape):
        assert len(
            input_shape) == 3, f"Number of dimensions should be 3, got {len(input_shape)}"
        super().build(input_shape)
        pe_init = keras.initializers.TruncatedNormal(stddev=0.02)
        self.pe = tf.Variable(
            name="pos_embedding",
            initial_value=pe_init(shape=(1, input_shape[1], input_shape[2])),
            dtype="float32",
            trainable=True,
        )

    def call(self, inputs, training=None):
        return inputs + tf.cast(self.pe, dtype=inputs.dtype)


@tf.keras.utils.register_keras_serializable()
class QuantizedClassToken(Calibrable, ClassToken):
    """Quantize the :class:`ClassToken` layer, allowing quantization of the output.

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        self.cls_quantizer = deserialize_quantizer(
            self.quant_config, "cls_quantizer", WeightQuantizer, True)
        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()
        self.cls_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QuantizedClassToken only accepts FixedPoint\
                              inputs. Receives {type(inputs)} inputs.")

        batch_size = tf.shape(inputs.values)[0]
        inputs = inputs.promote(self.buffer_bitwidth)

        cls = self.cls_quantizer(self.cls, training=training)
        cls = cls.promote(self.buffer_bitwidth)

        cls_broadcasted = tf.broadcast_to(cls, [batch_size, 1, self.hidden_size])

        # Align inputs and cls_brodcasted frac_bits values before concatenation
        inputs, shift = inputs.align(cls_broadcasted)
        self.input_shift(shift)
        cls_broadcasted, shift = cls_broadcasted.align(inputs)
        self.cls_shift(shift)

        outputs = tf.concat([cls_broadcasted, inputs], 1)

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs, training=training)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config


@tf.keras.utils.register_keras_serializable()
class QuantizedAddPositionEmbs(Calibrable, AddPositionEmbs):
    """Quantize the :class:`AddPositionEmbs` layer, allowing operations in FixedPoint domain.

    Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        self.pe_quantizer = deserialize_quantizer(
            self.quant_config, "pe_quantizer", WeightQuantizer, True)
        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()
        self.pe_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QuantizedAddPositionEmbs only accepts FixedPoint\
                              inputs. Receives {type(inputs)} inputs.")

        inputs = inputs.promote(self.buffer_bitwidth)
        pe = self.pe_quantizer(self.pe, training=training).promote(self.buffer_bitwidth)

        # Align intermediate outputs and biases before adding them
        inputs, shift = inputs.align(pe)
        self.input_shift(shift)
        pe, shift = pe.align(inputs)
        self.pe_shift(shift)
        outputs = inputs + pe

        if self.out_quantizer is not None:
            outputs = self.out_quantizer(outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config


@tf.keras.utils.register_keras_serializable()
class ExtractToken(keras.layers.Layer):
    """ Wrapper class of `tf.gather` operation that allows to extract a Token.

    Args:
        token (int): the indice of the token to extract.

    """

    def __init__(self, *args, token, **kwargs):
        super().__init__(*args, **kwargs)
        self.token = token

    def call(self, inputs, training=False):
        return tf.gather(inputs, self.token, axis=1)

    def get_config(self):
        config = super().get_config()
        config.update({"token": self.token})
        return config


@tf.keras.utils.register_keras_serializable()
class QuantizedExtractToken(ExtractToken):
    """ Quantized version of the ExtractToken layer. Accepts only FixedPoint inputs.
    """

    def call(self, inputs, training=False):
        if not isinstance(inputs, FixedPoint):
            # If the inputs is not a FixedPoint, raise an error
            raise TypeError(f"QuantizedExtractToken only accepts FixedPoint\
                              inputs. Receives {type(inputs)} inputs.")
        return tf.gather(inputs, self.token, axis=1)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import re
import warnings

from keras import Model

from .utils import deep_clone_model, insert_layer
from .transforms import invert_batchnorm_pooling, fold_batchnorms
from ..layers import (QuantizedDense, QuantizedLayerNormalization, QuantizedAddPositionEmbs,
                      QuantizedClassToken, QuantizedGELU, QuantizedReLU, QuantizedAdd, QDropout,
                      QuantizedReshape, QuantizedAttention, QuantizedConv2D, QuantizedMaxPool2D,
                      QuantizedGlobalAveragePooling2D, QuantizedExtractToken, Dequantizer,
                      Quantizer, QuantizedSeparableConv2D, QuantizedFlatten, QuantizedReciprocal,
                      OutputQuantizer, WeightQuantizer, QuantizedConv2DTranspose,
                      QuantizedShiftmax, QuantizedSeparableConv2DTranspose)

# Mapper to match float layer with its quantized version
CUSTOM_QLAYERS = {
    "Conv2D": QuantizedConv2D, "Dense": QuantizedDense, "Add": QuantizedAdd,
    "AddPositionEmbs": QuantizedAddPositionEmbs, "ClassToken": QuantizedClassToken,
    "GELU": QuantizedGELU, "ReLU": QuantizedReLU, "Dropout": QDropout,
    "LayerMadNormalization": QuantizedLayerNormalization, "Reshape": QuantizedReshape,
    "Attention": QuantizedAttention, 'MaxPooling2D': QuantizedMaxPool2D,
    'GlobalAveragePooling2D': QuantizedGlobalAveragePooling2D,
    "ExtractToken": QuantizedExtractToken, "SeparableConv2D": QuantizedSeparableConv2D,
    "Flatten": QuantizedFlatten, "Reciprocal": QuantizedReciprocal, "PaddedConv2D": QuantizedConv2D,
    "Conv2DTranspose": QuantizedConv2DTranspose, "Shiftmax": QuantizedShiftmax,
    "SeparableConv2DTranspose": QuantizedSeparableConv2DTranspose
}
# List of Quantizer layer's that do not have a float layer representation
NO_FLOAT_CUSTOM_QLAYERS = [Dequantizer, OutputQuantizer, WeightQuantizer]


def _handle_not_quantizable_layers(model):
    """ Checks if the model has not quantizable layers and adds a Dequantizer before.

    Args:
        model (keras.Model): model to check

    Returns:
        keras.Model: the updated model
    """
    def is_quantizable(layer):
        layer_class = layer.__class__
        return (layer_class.__name__ in CUSTOM_QLAYERS or layer_class in CUSTOM_QLAYERS.values()
                or layer_class in NO_FLOAT_CUSTOM_QLAYERS)

    # Find layers that cannot be quantized
    for layer in model.layers:
        if not is_quantizable(layer):
            # This layer cannot be quantized, check its inbounds
            inbound = layer.inbound_nodes[0]
            if not inbound.inbound_layers:
                # Skip input layers
                continue
            elif isinstance(inbound.inbound_layers, list):
                raise RuntimeError(f"'{layer.name}' is not quantizable and has multiple inbounds "
                                   "which is not supported.")

            # Check if the layer inbound will be quantized but is not a Dequantizer to prevent
            # adding an additional Dequantizer.
            if (is_quantizable(inbound.inbound_layers) and
                    not isinstance(inbound.inbound_layers, Dequantizer)):
                # Inbound will be quantized, add a Dequantizer after it and return the model
                inbound_name = inbound.inbound_layers.name
                model = insert_layer(model, inbound_name,
                                     Dequantizer(name=f"{inbound_name}_float_output"))
                warnings.warn(f"'{layer.name}' of type {layer.__class__} is not supported to "
                              "quantize, a Dequantizer is added before it and quantization will "
                              "stop at this layer.")
                return model
    return model


def quantize(model, q_config, add_dequantizer=True):
    """Quantizes the model using the provided configuration.

    Applying the quantization configuration is done in two steps, giving priority to the first:

    1. Match the quantization configuration from the layer name, as long as the final part of its
       name matches with the configuration key.
       Example: To quantize all "Transformer/EncoderBlock_x/MlpBlock/activation" layers, a
       configuration with the key "activation" or "MlpBlock/activation" will be enough. This allows
       the possibility to have a specific configuration for the activation function of block 5 using
       the "Transformer/EncoderBlock_5/MlpBlock/activation" key.

       Notes:

            * If configuration for all blocks are not defined, a global configuration is required.
            * To avoid ambiguity, the key with the most details in terms of hierarchy should be
              selected, e.g. "activation" key in the quantization configuration will be enough to
              quantize all "activation" layers. However, if it is required to define a different
              configuration in all MLP blocks activations, configuration key should be
              "MlpBlock/activation" instead.

    2. Match the quantization configuration from the layer name, as long as the format
       '.../ParentLayer<_i>/SubParentLayer<_j>/.../Layer<_k>' matches with layer.name, where {i, j,
       k} are optional integers. Example: "EncoderBlock/add" key will quantize all layers ending
       with the format "EncoderBlock_i/add_j" ("EncoderBlock_0/add_0", "EncoderBlock_0/add_1",
       "EncoderBlock_1/add_0", "EncoderBlock_1/add_1", ...).

        Note:

            The configuration priority in this case is defined by the number of hierarchical terms
            and the detail in subindices from the end to the beginning of the key name. That is
            "EncoderBlock/add_0" will have priority over "EncoderBlock/add" and
            "EncoderBlock_0/add". The "EncoderBlock_0/add_0" case is covered by the previous point.

    Args:
        model (keras.Model): the model to quantize
        q_config (dict): quantization configuration
        add_dequantizer (bool, optional): allows to convert output to float. Defaults to True.

    Returns:
        keras.Model: the quantized model
    """
    # Add "/" at the beginning of the names, to avoid ambiguity (e.g. "Dense" and "DropoutDense")
    q_config = {key if key[0] == "/" else "/" + key: value for key, value in q_config.items()}

    # This iterates through the <key, config> in the JSON, and creates a regexp based on the key.
    # The regexp is built by:
    #   - identifying items with the '/' separator,
    #   - replace trailing <_N> in each item by a matching pattern,
    #   - concatenate back items matching patterns.
    q_config_regex = {re.compile("/" + r"(?:|_[\d]+)/".join(key[1:].split('/')) + r"(?:|_[\d]+)$"):
                      (key[1:], obj) for key, obj in q_config.items()}

    # Layers will no longer be quantized as soon as a Dequantizer is found
    quantization_stopped = False

    def _replace_layer(layer):
        nonlocal quantization_stopped
        config = layer.get_config()

        # Function to handle unsupported arguments in config
        def pop_unsupported_args(class_type):
            for arg, default_value in getattr(class_type, "unsupported_args", {}).items():
                if (arg in config and config[arg] != default_value):
                    raise RuntimeError(
                        f"Argument '{arg}' in layer '{layer.name}' is only "
                        f"supported with default value '{default_value}'. "
                        f"Receives '{config[arg]}'.")
                config.pop(arg, None)

        # Function that return a quantized layer given its float version
        def get_quantize_layer(layer, quantize_config={}):
            """Quantize float layer in three steps:
                - first, we get its quantized version,
                - second, remove unsupported arguments,
                - then, we return the quantized layer with config updated
            """
            nonlocal quantization_stopped
            class_name = layer.__class__.__name__
            # 1.1 Check if qlayer exists in custom layers
            if class_name in CUSTOM_QLAYERS:
                qlayer = CUSTOM_QLAYERS[class_name]
            # 1.2 Else, return the float version of the layer
            else:
                qlayer = layer
                qlayer_class = qlayer.__class__
                if not (qlayer_class in CUSTOM_QLAYERS.values() or
                        qlayer_class in NO_FLOAT_CUSTOM_QLAYERS):
                    warnings.warn(
                        f"'{class_name}' is not supported to quantize. It will be ignored.")
                # If a Dequantizer is found, quantization must be stopped
                quantization_stopped = isinstance(layer, Dequantizer)

            # 2. Remove unsupported arguments
            pop_unsupported_args(qlayer)

            # 3. Instantiate quantized layer
            if len(quantize_config) > 0:
                config['quant_config'] = quantize_config
            return qlayer.from_config(config)

        # When a not quantizable layer is found, stop quantization returning initial layer
        if quantization_stopped:
            return layer.from_config(config)

        # Replace float layer by quantized layer in three different ways:
        # 1. Check if layer is in quant_config (to quantize specific layers)
        layer_name = layer.name if layer.name[0] == "/" else "/" + layer.name
        match_name = [key for key in q_config.keys() if layer_name.endswith(key)]
        if len(match_name) > 0:
            # In case of multiple matches, take the most hierarchical specific one
            match_name = sorted(match_name, key=lambda x: len(
                x.split('/')), reverse=True)[0]
            match_conf = q_config[match_name]
            return get_quantize_layer(layer, match_conf)

        # 2. Suppose that the layer name has the following format:
        #    .../ParentLayer<_i>/SubParentLayer<_j>/.../Layer<_k>, where ijk are integers and
        #    <_i>, <_j>, <_k> are optional.
        #    Match the pattern and replace the layer with a quantized version.
        match_tuples = []
        for rkey, qobj in q_config_regex.items():
            if rkey.search(layer_name):
                match_tuples.append(qobj)
        if len(match_tuples) > 0:
            # In case of multiple matches, takes the configurations with the highest hierarchy
            max_hierarchy = max([len(qobj[0].split('/'))
                                for qobj in match_tuples])
            match_tuples = {tuple(qobj[0].split('/')[::-1]): qobj[1]
                            for qobj in match_tuples if len(qobj[0].split('/')) == max_hierarchy}
            # And then, chooses the configuration with the most detailed sub-indices,
            # from the end to the beginning.
            #
            # For example, if the layer name is: A_1/B_2/C_3/D_4, and the config keys are:
            # keys = [B_2/C/D, B/C_3/D], the code will search the first exact match from bottom
            # to top, following:
            # First step (layer_name[0], keys[0]): if D_4 == D ? No, go to next step
            # Second step (layer_name[0], keys[1]): if D_4 == D ? No, go to next step
            # Third step (layer_name[1], keys[0]): if C_3 == C ? No, go to next step
            # Fourth step (layer_name[1], keys[1]): if C_3 == C_3 ? Yes, match with config[key[1]]
            #
            # Note: If only exist this keys in config file, keys[0] will be used in the following
            # layers: A_i/B_2/C_j/D_k, for all ijk in [0, ..., n] and j != 3.
            for (idx, pattern) in enumerate(layer_name[1:].split('/')[::-1]):
                match_tuples_copy = {}
                for (key, match_conf) in match_tuples.items():
                    if key[idx] == pattern:
                        match_tuples_copy[key] = match_conf
                # If only one match is found, it is used as the configuration
                if len(match_tuples_copy) == 1:
                    match_conf = list(match_tuples_copy.values())[0]
                    break
                # If length of keys is smaller than hierarchical name, use the last match_conf
                if idx + 1 >= max_hierarchy:
                    break
                # Update match_tuples only when patterns match, in case of a tie
                if len(match_tuples_copy) > 1:
                    match_tuples = match_tuples_copy
            return get_quantize_layer(layer, match_conf)

        # 3. Special cases for Embedding layers
        #    Cases covered in previous steps

        # If no match, return the qlayer with config unchanged
        return get_quantize_layer(layer)

    # Invert BN <-> Pooling layers and fold BN into their preceding layers
    model = invert_batchnorm_pooling(model)
    model = fold_batchnorms(model)

    # Check if the model has not quantizable layers and add a Dequantizer before
    model = _handle_not_quantizable_layers(model)

    # Quantize the model replacing layers with their quantized version
    new_model = deep_clone_model(model, clone_function=_replace_layer)
    out = new_model.outputs

    # Append Dequantizer at the end of the model to convert the output to float value
    if (add_dequantizer and not isinstance(new_model.layers[-1], Dequantizer)
            and not quantization_stopped):
        out = Dequantizer(name="float_output")(out)
    return Model(new_model.input, out)


def dump_config(model, skip_default=False):
    """Dump the quantization configuration of a quantized model, exporting the configuration for
    each quantized layer.

    Args:
        model (keras.Model): a quantized model.
        skip_default (bool, optional): remove default values on each quantizer. Defaults to False.

    Returns:
        dict: the configuration of the model.
    """
    # Get the simplest configuration by default of one quantizer
    dump_obj = {"signed": True, "axis": "per-tensor", "bitwidth": None, "scale_bits": None}

    # Function that update the configuration of each quantizer in one layer, reducing the
    # parameters if short_form is True, otherwise adding the default parameters.
    def _update_config(config):
        for conf_value in config.values():
            if isinstance(conf_value, dict):
                # Check if current config is a singular quantizer
                try:
                    scale_bits = conf_value.pop("scale_bits", None)
                    Quantizer(**conf_value)
                    if scale_bits is not None:
                        conf_value["scale_bits"] = scale_bits
                except Exception:
                    _update_config(conf_value)
                    continue

                # Remove extra keys
                keys_to_pop = [key for key in conf_value.keys() if key not in dump_obj]
                for ckey in keys_to_pop:
                    conf_value.pop(ckey)

                # Update the configuration of a singular quantizer follows:
                # * skip_default=True: removes default values, given by Quantizer.get_config()
                # * skip_default=False: keeps only [bitwidth, signed, axis, scale_bits] parameters
                for dkey, dval in dump_obj.items():
                    if not skip_default and dval is not None and dkey not in conf_value:
                        conf_value[dkey] = dval
                    elif skip_default and dval == conf_value.get(dkey, -1):
                        conf_value.pop(dkey)

    # Get the configuration of the model, iterating over each layer and updating on config.
    config = {}
    for layer in model.layers:
        # Try to take the current quantized configuration
        ly_config = layer.get_config().get('quant_config')

        # Only append quantized configuration
        if layer.__class__ in CUSTOM_QLAYERS.values() and ly_config is not None:
            _update_config(ly_config)
            config[layer.name] = ly_config

    return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from copy import copy
import tensorflow as tf


from ..debugging import assert_equal


@tf.function
@tf.custom_gradient
def round_through(x):
    rounded = tf.math.round(x)

    def grad(upstream):
        return upstream
    return rounded, grad


@tf.function
@tf.custom_gradient
def floor_through(x):
    floored = tf.math.floor(x)

    def grad(upstream):
        return upstream
    return floored, grad


@tf.function
@tf.custom_gradient
def ceil_through(x):
    ceiled = tf.math.ceil(x)

    def grad(upstream):
        return upstream
    return ceiled, grad


class QTensor(tf.experimental.ExtensionType):
    """Abstract class to exchange quantized tensors between layers

    The QTensor values are actually stored as integer, but it provides a
    conversion method to project these values into a float representation.

    The value_bits parameter sets the maximum integer values that can be stored:

        int_max = 2^bits - 1.

    When a QTensor is created, its values are clipped to [-int_max-1, int_max].
    """
    values: tf.Tensor = 1.0
    value_bits: int = 7
    shape: tf.TensorShape  # Required to convert to a KerasTensor

    @property
    def per_tensor(self):
        """Returns if QTensor is quantized per-tensor

        Returns:
            bool: True if QTensor is quantized per-tensor or False on per-axis case.
        """
        raise NotImplementedError

    def to_float(self):
        """Returns a float representation of the QTensor

        Returns:
            :obj:`tensorflow.Tensor`: the float representation.
        """
        raise NotImplementedError

    def clone(self):
        """Returns a copy of the QTensor

        Returns:
            :obj:`QTensor`: the copy.
        """
        return copy(self)

    def __str__(self):
        x_float = self.to_float()
        return f"QTensor: {x_float}"

    @staticmethod
    def int_max(value_bits):
        return 2 ** value_bits - 1

    @staticmethod
    def clamp(values, value_bits):
        int_max = QTensor.int_max(value_bits)
        return tf.clip_by_value(values, -int_max - 1, int_max)

    def assert_per_tensor(self):
        """Asserts that a QTensor is quantized per-tensor"""
        name = self.__class__.__name__ if not hasattr(self.values, "name") else self.values.name
        assert_equal(self.per_tensor, True, message=f"{name} is not per-tensor.")


def pow2(n):
    """Return the power of two of an integer

    Note that this goes through a float64 operation to obtain a float32 tensor.
    This makes sure the operation can run on a GPU without losing precision.

    Args:
        n (`tf.tensor`, int): the positive or negative exponent

    Returns:
        :obj:`tensorflow.Tensor`: a float tensor containing the PoT of the input.
    """
    return tf.cast(2.0 ** tf.cast(n, tf.float64), tf.float32)


def ceil_log2(x):
    """Return the closest power of two exponent of a float tensor.

    This evaluates for each element of the input tensor the integer exponent leading
    to the closest power-of-two higher than the input.

    In hardware, if the inputs are represented as integer, this operation can
    be implemented by identifying the leading bit index and increment the result
    by 1.

    Example: ceil_log2(7) = ceil_log2(0b00000111) = 2 + 1 = 3

    Args:
        x (:obj:`tensorflow.tensor`): the source tensor

    Returns:
        :obj:`tensorflow.Tensor`: a float tensor containing integer values
            representing the closest PoT exponents.
    """
    return ceil_through(tf.experimental.numpy.log2(tf.cast(x, tf.float32)))


def floor_log2(x):
    """Return the closest power of two exponent of a float tensor.

    This evaluates for each element of the input tensor the integer exponent leading
    to the closest power-of-two lower than the input (except if x < 1).

    In hardware, if the inputs are represented as integer, this operation can
    be implemented by identifying the leading bit index.

    Example: floor_log2(7) = floor_pow2(0b00000111) = 2

    Args:
        x (:obj:`tensorflow.tensor`): the source tensor

    Returns:
        :obj:`tensorflow.Tensor`: a float tensor containing integer values
            representing the closest PoT exponents.
    """
    return floor_through(tf.experimental.numpy.log2(tf.cast(x, tf.float32)))


def round_log2(x):
    """Return the closest power of two exponent of a float tensor.

    This evaluates for each element of the input tensor the integer exponent leading
    to the closest power-of-two.

    In hardware, if the inputs are represented as integer, this operation can
    be implemented by:
    - identifying the leading bit index,
    - increment by 1 if the previous bit is 1 also.

    Example: round_log2(7) = round_log2(0b00000111) = 2 + 1 = 3
             round_log2(5) = round_log2(0b00000101) = 2

    Args:
        x (:obj:`tensorflow.tensor`): the source tensor

    Returns:
        :obj:`tensorflow.Tensor`: a float tensor containing integer values
            representing the closest PoT exponents.
    """
    return round_through(tf.experimental.numpy.log2(tf.cast(x, tf.float32)))



from os import path
from setuptools import setup, find_packages
from platform import platform


def get_tf_dep():
  platform_string = platform()
  if 'macOS' in platform_string and 'arm64' in platform_string:
    tf_name = 'tensorflow-macos'
  else:
    tf_name = 'tensorflow'
  return tf_name + '>=2.10.0'


# Read the contents of the README file
directory = path.abspath(path.dirname(__file__))
with open(path.join(directory, 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='quantizeml',
      version='0.0.12',
      description='Base layers and quantization tools',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='David Corvoysier',
      author_email='dcorvoysier@brainchip.com',
      url='https://doc.brainchipinc.com',
      license='Apache 2.0',
      license_files=['LICENSE'],
      packages=find_packages(),
      entry_points={
        'console_scripts': [
            'quantizeml = quantizeml.cli:main',
        ]
      },
      install_requires=[get_tf_dep(), 'keras>=2.10.0', 'tensorflow_addons>=0.18.0'],
      python_requires='>=3.7')



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import tensorflow as tf

from quantizeml.layers import WeightQuantizer, OutputQuantizer, QuantizedReLU
from quantizeml.tensors import QTensor, QFloat, ceil_log2, floor_log2


def check_quantization(model):
    """Checks the specified model quantization.

    It looks for errors that can be fixed in the quantization configuration:

    - inaccurate weight scales quantization,
    - saturation in integer operations.

    Args:
        model (keras.Model): the model to check

    Returns:
        list(str): the quantization issues that were detected
    """
    messages = []
    for layer in model.layers:
        for name, attr in layer.__dict__.items():
            if isinstance(attr, WeightQuantizer):
                quantizer = attr
                if name == "bias_quantizer":
                    w = layer.bias
                elif name == "dw_weight_quantizer":
                    w = layer.depthwise_kernel
                elif name == "pw_weight_quantizer":
                    w = layer.pointwise_kernel
                elif name == "beta_quantizer":
                    w = layer.get_weights()[1]
                else:
                    w = layer.get_weights()[0]
                w_q = quantizer(w)
                if isinstance(w_q, QFloat):
                    axis = list(range(len(w.shape) - 1))
                    max_values = tf.reduce_max(tf.abs(w), axis)
                    ideal_scales = max_values / QTensor.int_max(quantizer.value_bits)
                    err = tf.abs(ideal_scales - w_q.scales.to_float()) / tf.abs(ideal_scales)
                    mean_err = tf.reduce_mean(err)
                    if mean_err > 5e-2:
                        message = f"Scales quantization relative error is high in " \
                                    f"{layer.name}/{quantizer.name}: {mean_err:.4f}."
                        if quantizer._axis == "per-tensor":
                            message += "Use a per-axis quantizer and/or increase scales bits."
                        else:
                            message += "Try increasing scales bits."
                        messages.append(message)
            elif isinstance(attr, OutputQuantizer):
                quantizer = attr
                max_value = quantizer.get_weights()[0]
                # Since output quantizer max_values are evaluated by dividing the FixedPoint
                # integer values by their PoT scales, the result can only be a PoT (positive
                # or negative) if the initial value was a PoT. This is very unlikely to
                # happen except if the previous operation saturated its output, resulting in
                # a clip to a PoT.
                pot_max_value = ceil_log2(max_value) == floor_log2(max_value)
                if isinstance(layer, QuantizedReLU):
                    # Remove false positives clipped to zero
                    pot_max_value = tf.math.logical_and(max_value != 0., pot_max_value)
                    # Remove false positives clipped to the ReLU max_value
                    pot_max_value = tf.math.logical_and(max_value != layer.max_value, pot_max_value)
                if tf.reduce_any(pot_max_value):
                    messages.append(f"Saturation detected before {layer.name}/{quantizer.name}: "
                                    "check the previous operation and/or increase buffer bitwidth.")
    return messages



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
import keras

from ...tensors import FixedPoint


__all__ = ["QuantizedFlatten"]


@tf.keras.utils.register_keras_serializable()
class QuantizedFlatten(keras.layers.Flatten):
    """A Flatten layer that operates on quantized inputs
    """

    def call(self, inputs, training=None):
        # raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError("QuantizedFlatten only accepts FixedPoint inputs. "
                            f"Receives {type(inputs)} inputs.")
        if inputs.frac_bits.shape.ndims:
            raise ValueError("QuantizedFlatten only supports inputs quantized per-tensor.")
        # Flatten the values
        flattened_values = super().call(inputs.values)
        # Return a new reshaped FixedPoint
        return FixedPoint(flattened_values, inputs.frac_bits, inputs.value_bits)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
from typing import List

from ..fixed_point import FixedPoint


@tf.experimental.dispatch_for_api(tf.shape)
def fp_shape(input: FixedPoint, out_type=tf.dtypes.int32, name=None):
    return tf.shape(input.values, out_type, name)


@tf.experimental.dispatch_for_api(tf.reshape)
def fp_reshape(tensor: FixedPoint, shape, name=None):
    tensor.assert_per_tensor()
    output = tf.reshape(tensor.values, shape, name)
    return FixedPoint(output, tensor.frac_bits, tensor.value_bits)


@tf.experimental.dispatch_for_api(tf.transpose)
def fp_transpose(a: FixedPoint, perm=None, conjugate=False, name="transpose"):
    a.assert_per_tensor()
    output = tf.transpose(a.values, perm, conjugate, name)
    return FixedPoint(output, a.frac_bits, a.value_bits)


@tf.experimental.dispatch_for_api(tf.broadcast_to)
def fp_brodcast_to(input: FixedPoint, shape, name=None):
    """Broadcast a FixedPoint tensor for a compatible shape.

    Args:
        input (:obj:`FixedPoint`): a FixedPoint to broadcast.
        shape (:obj:`tf.Tensor`): an 1-D `int` Tensor representing
            the shape of the desired output. Must be one of the
            following types: `int32`, `int64`.
        name (str, optional): a name for the operation. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the brodcasted output. Has the same
            type as `input`.
  """
    # Check first that the last dimension is unchanged
    tf.assert_equal(input.shape[-1], shape[-1], message="To brodcast FixedPoint input,\
                        last dimension should remain unchanged")
    output = tf.broadcast_to(input.values, shape, name)
    return FixedPoint(output, input.frac_bits, input.value_bits)


@tf.experimental.dispatch_for_api(tf.concat)
def fp_concat(values: List[FixedPoint], axis, name="concat"):
    """Concatenates FixedPoint tensors along one dimension.

    Args:
        values (List of :obj:`FixedPoint`): List of FixedPoint tensors
            to concatenate.
        axis (list): Dimension along which to concatenate.
        name (str, optional): the name for the Tensorflow ops.
            Defaults to "concat".

    Returns:
        :obj:`FixedPoint`: the concatenate output FixedPoint.
    """
    if len(values) == 1:
        return FixedPoint(values[0].values, values[0].frac_bits, values[0].value_bits)

    # For now we only support concatenation of one or two elements
    tf.assert_equal(
        len(values),
        2, f"We only support concatenation of one or two FixedPoint. \
           Receives {len(values)} tensors as input.")

    # For now we don't support concatenation over last dimension
    rank_input = tf.rank(values[0].values)
    if axis < 0:
        dim = axis + rank_input
    else:
        dim = axis
    tf.Assert(tf.math.not_equal(dim, rank_input - 1),
              ['we do not support concatenation over last axis. axis is:', axis])

    # FixedPoint tensors to concatenate should be per-tensor
    values[0].assert_per_tensor()
    values[1].assert_per_tensor()

    tf.assert_equal(values[0].frac_bits, values[1].frac_bits, message=f"the two\
            FixedPoint must have the same frac_bits. Receives {values[0].frac_bits}\
            and {values[1].frac_bits}")
    values_out = tf.concat([values[0].values, values[1].values], axis, name)

    return FixedPoint(
        values_out, values[0].frac_bits, max(
            values[0].value_bits, values[1].value_bits))


@tf.experimental.dispatch_for_api(tf.expand_dims)
def fp_expand_dims(input: FixedPoint, axis, name=None):
    """Returns a tensor with a length 1 axis inserted at index `axis`.

    Args:
        input (FixedPoint): a `Tensor`.
        axis (int): integer specifying the dimension index at which to expand the shape of `input`.
            Given an input of D dimensions, `axis` must be in range `[-(D+1), D]` (inclusive).
        name (str, optional): name of the output `Tensor`. Defaults to None.

    Returns:
        FixedPoint: a tensor with the same data as `input`, with an additional dimension inserted at
            the index specified by `axis`.
    """
    # Only support per-tensor inputs
    input.assert_per_tensor()

    # Expand dimension on values
    values = tf.expand_dims(input.values, axis, name)

    # Return a new FixedPoint
    return FixedPoint(values, input.frac_bits, input.value_bits)


@tf.experimental.dispatch_for_api(tf.compat.v1.gather)
def fp_gather(params: FixedPoint, indices, validate_indices=None,
              name=None, axis=None, batch_dims=0):
    """Gather slices from params along axis.

    Args:
        params (:obj:`FixedPoint`, :obj:`QFloat`): the input QTensor.
        indices (int, list): the indices to gather.
        validate_indices (bool, optional): whether to validate the indices. Defaults to None.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.
        axis (int, optional): the axis to gather along. Defaults to None.
        batch_dims (int, optional): the number of batch dimensions to keep. Defaults to 0.

    Returns:
        :obj:`QTensor`: a QTensor containing the gathered values.
    """
    # We do not support gather along the last axis if the QTensor is per-axis
    if axis in (-1, len(params.shape)):
        params.assert_per_tensor()
    # Create a new QTensor, gather the indices in a desired axis
    x_gather = tf.gather(params.values, indices, validate_indices, axis, batch_dims, name)
    return FixedPoint(x_gather, params.frac_bits, params.value_bits)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from keras.layers import MaxPool2D, GlobalAveragePooling2D

from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, OutputQuantizer
from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH

__all__ = ["QuantizedMaxPool2D", "QuantizedGlobalAveragePooling2D"]


@tf.keras.utils.register_keras_serializable()
class QuantizedMaxPool2D(MaxPool2D):
    """A max pooling layer that operates on quantized inputs.

    """

    def call(self, inputs):
        # Raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QuantizedMaxPool2D only accepts FixedPoint inputs. \
                             Receives {type(inputs)} inputs.")

        outputs = super().call(inputs.values)
        return FixedPoint(outputs, inputs.frac_bits, inputs.value_bits)


@tf.keras.utils.register_keras_serializable()
class QuantizedGlobalAveragePooling2D(Calibrable, GlobalAveragePooling2D):
    """A global average pooling layer that operates on quantized inputs.

     Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """

    def __init__(self, quant_config={}, **kwargs):
        super().__init__(**kwargs)
        self.quant_config = quant_config
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        self.buffer_bitwidth = self.quant_config.get("buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()

    def build(self, input_shape):
        super().build(input_shape)
        # Build the spatial size
        self.spatial_size = (input_shape[1] * input_shape[2])

    def call(self, inputs, training=None):
        # Raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QuantizedGlobalAveragePooling2D only accepts FixedPoint inputs. \
                             Receives {type(inputs)} inputs.")

        # Promote input to prevent overflow in the reduce_sum op, then align them
        inputs, shift = inputs.promote(self.buffer_bitwidth).align()
        self.input_shift(shift)
        inputs_sum = tf.reduce_sum(inputs, axis=[1, 2], keepdims=self.keepdims)
        inputs_mean = inputs_sum / self.spatial_size

        if self.out_quantizer is not None:
            return self.out_quantizer(inputs_mean, training=training)
        return inputs_mean

    def get_config(self):
        config = super().get_config()
        config.update({"quant_config": self.quant_config})
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
import keras
from tensorflow.python.framework import ops

from .calibrable import Calibrable, CalibrableVariable
from .quantizers import deserialize_quantizer, WeightQuantizer, OutputQuantizer
from ..tensors import FixedPoint, MAX_BUFFER_BITWIDTH


__all__ = ["QuantizedLayerNormalization", "LayerMadNormalization"]


@tf.keras.utils.register_keras_serializable()
class LayerMadNormalization(keras.layers.LayerNormalization):
    r"""Approximates the :obj:`keras.layers.LayerNormalization` (LN), replacing the computation of
    the standard deviation by the mean average deviation (mad).

    Taking into account the complexity of the calculate required in the standard deviation,
    the LayerMadNormalization (LMN) is intended to replace the :math:`std(x)` by :math:`mad(x)`
    on a Power of Two (PoT), defined as:

    .. math:: mad(x) = |x - mean(x)| * 2^\text{-nb_channels}

    To simplify it even more and make it more hardware-friendly, :math:`mean(x)` is considered
    as 0, so. These are the resulting calculations:

    .. math:: mad(x) = sum(|x|) * 2^\text{-nb_channels}

    Then, the equation of the layer is defined as:

    .. math:: LMN(x) = \gamma\frac{x}{mad(x)} + \beta

    .. note::
        A tuning step in the switching procedure between the LN to LMN layer will be required
        to find the :math:`(\gamma, \beta)` parameters that match the standard deviation changes.
    """

    def build(self, input_shape):
        super().build(input_shape)

        # Save an approximation of nb_channels as PoT to calculate the mad of inputs as:
        # mad = sum(abs(inputs)) * 2 ** -mad_shift
        if len(self.axis) > 1:
            raise ValueError(f"{self.__class__.__name__} does not support multiple axis.")
        nb_channels = tf.cast(input_shape[self.axis[0]], tf.float32)
        self.mad_shift = tf.math.round(tf.math.log(nb_channels) / tf.math.log(2.))

    def _get_mad(self, inputs, axis, *args, name=None, **kwargs):
        """Compute the mean absolute deviation (mad).

        Args:
            inputs (:obj:`tensorflow.Tensor`): input tensor.
            axis (array of ints): axis along which to compute mad.
            name (str, optional): name used to scope the operations that compute the mad.
                Defaults to None.
            *args, **kwargs (dict, optional): additional arguments of :func:`tf.math.reduce_sum`.

        Returns:
            :obj:`tensorflow.Tensor`: mean absolute deviation (mad).
        """
        # Compute the mad, taking into account the approximations:
        # mad \approx sum(abs(inputs))
        # nb_channels \approx 2 ** round(log2(inputs.shape[axis]))
        nb_channels_inv = 2.0 ** -self.mad_shift
        with ops.name_scope(name, "_mad", [inputs, axis]):
            mad = tf.math.reduce_sum(tf.math.abs(inputs), axis, *args, **kwargs, name="mad")
            mad = mad * nb_channels_inv
        return mad

    def call(self, inputs, training=None):
        """Use the same call as the parent, changing tf.nn.moments instead of self._get_mad,
        and replacing mean by 0"""
        # Compute the axis along which to reduce the mad
        input_shape = inputs.shape
        ndims = len(input_shape)

        # Broadcasting only necessary for norm when the axis is not just the last dimension
        broadcast_shape = [1] * ndims
        for dim in self.axis:
            broadcast_shape[dim] = input_shape.dims[dim].value

        def _broadcast(v):
            if (v is not None and len(v.shape) != ndims and self.axis != [ndims - 1]):
                return tf.reshape(v, broadcast_shape)
            return v

        input_dtype = inputs.dtype
        if input_dtype in ('float16', 'bfloat16') and self.dtype == 'float32':
            # If mixed precision is used, cast inputs to float32 so that this is at
            # least as numerically stable as the fused version.
            inputs = tf.cast(inputs, 'float32')

        # Calculate the mad on the last axis (layer activations).
        mad = self._get_mad(inputs, self.axis, keepdims=True)
        scale, offset = _broadcast(self.gamma), _broadcast(self.beta)

        # Compute layer normalization as layer mad normalization
        outputs = scale * inputs / mad + offset
        outputs = tf.cast(outputs, input_dtype)

        # If some components of the shape got lost due to adjustments, fix that.
        outputs.set_shape(input_shape)

        return outputs


@tf.keras.utils.register_keras_serializable()
class QuantizedLayerNormalization(Calibrable, LayerMadNormalization):
    """ A LayerNormalization layer that operates on quantized inputs and weights.

     Args:
        quant_config (dict, optional): the serialized quantization configuration.
            Defaults to empty configuration.
    """
    def __init__(self, *args, quant_config={}, **kwargs):
        super().__init__(*args, **kwargs)
        self.quant_config = quant_config
        self.out_quantizer = deserialize_quantizer(
            self.quant_config, "output_quantizer", OutputQuantizer, False)
        self.buffer_bitwidth = self.quant_config.get(
            "buffer_bitwidth", MAX_BUFFER_BITWIDTH) - 1
        self.mad_rec_frac_bits = self.quant_config.get("mad_rec_frac_bits", 16)
        assert self.buffer_bitwidth > 0, "The buffer_bitwidth must be a strictly positive integer."
        assert self.mad_rec_frac_bits >= 0, "The mad_rec_frac_bits must be a non-negative integer."
        self.gamma_quantizer = deserialize_quantizer(
            self.quant_config, "gamma_quantizer", WeightQuantizer, True)
        self.beta_quantizer = deserialize_quantizer(
            self.quant_config, "beta_quantizer", WeightQuantizer, True)

        # Add objects that will store the shift values.
        self.input_shift = CalibrableVariable()
        self.x_rescaled_shift = CalibrableVariable()
        self.one_bits = CalibrableVariable()
        self.beta_shift = CalibrableVariable()
        self.output_shift = CalibrableVariable()

    def call(self, inputs, training=None):
        """The quantized version of the LayerNormalization with Integer-only operations.

            This normalizes the input tensor then returns a quantized output.

            inputs (:obj:`FixedPoint`): the inputs tensor.
            training (bool, optional): the training mode. Defaults to None.

            Returns:
                :obj:`FixedPoint`: output tensor.
        """
        # raise an error if the inputs are not FixedPoint
        if not isinstance(inputs, FixedPoint):
            raise TypeError(f"QuantizedLayerNormalization only accepts FixedPoint\
                              inputs. Receives {type(inputs)} inputs.")

        # Increase the value_bits of inputs to avoid saturation, then align them
        inputs, shift = inputs.promote(self.buffer_bitwidth).align()
        self.input_shift(shift)

        # Compute the Absolute Deviation (MAD)
        mad_sum = tf.math.reduce_sum(
            inputs.abs(), self.axis, keepdims=True, name="mad")
        # Calculate a gamma_rescaled = gamma * 2 ** mad_shift. This is the
        # value that will be quantized.
        gamma_rescaled = self.gamma * (2 ** self.mad_shift)
        # Quantize the learnable parameters gamma_rescaled and beta
        gamma_rescaled = self.gamma_quantizer(
            gamma_rescaled, training=training)
        beta = self.beta_quantizer(self.beta, training=training)

        # Calculate the gamma_on_mad_sum = (gamma * 2**mad_shift) / mad_sum.
        # This can be evaluated once on hardware.
        gamma_on_mad_sum = gamma_rescaled / mad_sum

        # Multiply inputs by gamma
        x_rescaled = inputs * gamma_on_mad_sum

        # Downscale to recover the initial fractional bits
        outputs, shift = x_rescaled.downscale(
            inputs.frac_bits, self.buffer_bitwidth)
        self.x_rescaled_shift(shift)

        # Before adding the bias, we need aligned results
        outputs, shift = outputs.align(beta)
        self.output_shift(shift)
        beta, shift = beta.promote(self.buffer_bitwidth).align(outputs)
        self.beta_shift(shift)
        outputs = tf.add(outputs, beta)
        outputs.values.set_shape(inputs.shape)

        # If no quantizer output is provided return the FixedPoint output tensor directly
        if self.out_quantizer is None:
            return outputs

        return self.out_quantizer(outputs, training=training)

    def get_config(self):
        config = super().get_config()
        config["quant_config"] = self.quant_config
        return config
